from __future__ import annotations
from typing import Any, Dict, List, Optional
from pydantic import BaseModel, Field
from .context import ContextObject

Json = Dict[str, Any]

class BusEvent(BaseModel):
    event_id: str
    topic: str
    type: str  # TASK | THOUGHT | RESULT | ALERT
    context: ContextObject = Field(default_factory=ContextObject)

    parent_id: Optional[str] = None
    sender: Optional[str] = None
    intended_for: List[str] = Field(default_factory=list)
    tags: List[str] = Field(default_factory=list)

    confidence: Optional[float] = None
    priority: int = 5
    meta: Json = Field(default_factory=dict)



from __future__ import annotations
import os, json
from typing import List
from .openapi_loader import load_openapi_operations
from .registry import ToolRegistry, ToolSpec

def discover_openapi_tools(directory: str, registry: ToolRegistry) -> List[str]:
    added: List[str] = []
    for root, _, files in os.walk(directory):
        for f in files:
            if not f.lower().endswith(".json"):
                continue
            path = os.path.join(root, f)
            with open(path, "r", encoding="utf-8") as fp:
                spec = json.load(fp)
            for op in load_openapi_operations(spec):
                registry.register(
                    ToolSpec(
                        name=op.tool_name,
                        description=op.description,
                        input_schema=op.input_schema,
                        output_schema=op.output_schema,
                        tags=["openapi", op.method.lower()],
                    ),
                    fn=_unbound_tool(op.tool_name),
                )
                added.append(op.tool_name)
    return added

def _unbound_tool(name: str):
    async def _fn(**kwargs):
        raise RuntimeError(
            f"Tool '{name}' discovered from OpenAPI but not bound to an executor. "
            f"Bind via your HTTP client adapter."
        )
    return _fn



from __future__ import annotations
from contextlib import contextmanager
from typing import Optional

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor

try:
    from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter  # type: ignore
except Exception:  # pragma: no cover
    OTLPSpanExporter = None  # type: ignore

_initialized = False

def init_tracing(otlp_endpoint: Optional[str] = None) -> None:
    global _initialized
    if _initialized:
        return
    provider = TracerProvider()
    if otlp_endpoint and OTLPSpanExporter is not None:
        exporter = OTLPSpanExporter(endpoint=otlp_endpoint)
        provider.add_span_processor(SimpleSpanProcessor(exporter))
    trace.set_tracer_provider(provider)
    _initialized = True

@contextmanager
def span(name: str):
    tracer = trace.get_tracer("adi-agentic-agi")
    with tracer.start_as_current_span(name):
        yield



from __future__ import annotations
from typing import Optional, Protocol, List
from pydantic import BaseModel
from ..core.events import BusEvent

class Checkpoint(BaseModel):
    checkpoint_id: str
    ts: float
    notes: str = ""

class EventStore(Protocol):
    async def put_event(self, event: BusEvent) -> None: ...
    async def get_event(self, event_id: str) -> Optional[BusEvent]: ...
    async def list_recent(self, limit: int = 100) -> List[BusEvent]: ...
    async def checkpoint(self, notes: str = "") -> Checkpoint: ...

class InMemoryEventStore:
    def __init__(self):
        self._events: dict[str, BusEvent] = {}
        self._order: list[str] = []

    async def put_event(self, event: BusEvent) -> None:
        if event.event_id not in self._events:
            self._order.append(event.event_id)
        self._events[event.event_id] = event

    async def get_event(self, event_id: str) -> Optional[BusEvent]:
        return self._events.get(event_id)

    async def list_recent(self, limit: int = 100) -> List[BusEvent]:
        ids = self._order[-limit:]
        return [self._events[i] for i in ids]

    async def checkpoint(self, notes: str = "") -> Checkpoint:
        import time, uuid
        return Checkpoint(checkpoint_id=str(uuid.uuid4()), ts=time.time(), notes=notes)



from __future__ import annotations
from typing import Any, Dict, List, Optional
from pydantic import BaseModel, Field

Json = Dict[str, Any]

class MediaBlob(BaseModel):
    mime_type: str
    data_b64: str
    meta: Json = Field(default_factory=dict)

class ContextObject(BaseModel):
    text: Optional[str] = None
    images: List[MediaBlob] = Field(default_factory=list)
    audio: List[MediaBlob] = Field(default_factory=list)
    data: Json = Field(default_factory=dict)
    citations: List[Json] = Field(default_factory=list)

    def merged(self, other: "ContextObject") -> "ContextObject":
        return ContextObject(
            text="\n".join([t for t in [self.text, other.text] if t]),
            images=[*self.images, *other.images],
            audio=[*self.audio, *other.audio],
            data={**self.data, **other.data},
            citations=[*self.citations, *other.citations],
        )



from __future__ import annotations
from typing import Optional, Set

from ..core.agent import BaseAgent
from ..core.decorators import agent_role
from ..core.events import BusEvent
from ..tools.python_sandbox import PythonSandbox
from ..tools.container_sandbox import DockerPythonSandbox

@agent_role("tool_smith")
class ToolSmithAgent(BaseAgent):
    def topics(self) -> Optional[Set[str]]:
        return {"task.code", "task.refine"}

    def filter_event(self, event: BusEvent) -> bool:
        return event.topic in {"task.code", "task.refine"} and event.type in {"TASK", "RESULT"}

    async def handle(self, event: BusEvent) -> None:
        await self.emit_confidence(parent=event, score=0.75, rationale="tool_smith ready")

        use_container = bool(self.config.extra.get("tool_smith_container", False))
        code_example = """result = 2 + 2
"""

        if use_container:
            sandbox = DockerPythonSandbox()
            res = await sandbox.run(code_example)
            out = f"container_ok={res.ok} exit={res.exit_code} stdout={res.stdout!r} stderr={res.stderr!r}"
        else:
            sandbox = PythonSandbox()
            res = await sandbox.run(code_example)
            out = f"sandbox_ok={res.ok} result={res.result_repr} stderr={res.stderr!r}"

        draft = (
            "### Implementation Draft\n"
            "- Execute code in a sandbox (local or container)\n"
            f"- Sandbox smoke test: {out}\n"
            "- Add module `solution.py`\n"
            "- Add tests `test_solution.py`\n"
            "- Add docs + type hints\n"
        )
        await self.report(parent=event, topic="result.draft", text=draft, confidence=0.75)



from __future__ import annotations
from typing import Optional, Set
from ..core.agent import BaseAgent
from ..core.decorators import agent_role
from ..core.events import BusEvent

@agent_role("context_librarian")
class ContextLibrarianAgent(BaseAgent):
    def topics(self) -> Optional[Set[str]]:
        return {"task.research"}

    def filter_event(self, event: BusEvent) -> bool:
        return event.topic == "task.research" and event.type == "TASK"

    async def handle(self, event: BusEvent) -> None:
        await self.emit_confidence(parent=event, score=0.6, rationale="rag stub")
        await self.report(parent=event, topic="result.research",
                          text="(RAG stub) Gather constraints, known pitfalls, and edge cases for the goal.", confidence=0.65)



from __future__ import annotations
from typing import Any, List

from .base import ProviderConfig, LLMMessage, LLMResponse

class GeminiProvider:
    """Google GenAI (Gemini) Provider (real implementation) using `google-genai` SDK."""
    def __init__(self, config: ProviderConfig):
        self.config = config
        try:
            from google import genai  # type: ignore
            from google.genai import types  # type: ignore
        except Exception as e:  # pragma: no cover
            raise RuntimeError("Install provider extra: pip install adi-agentic-agi[gemini]") from e

        self._genai = genai
        self._types = types

        # Gemini Developer API: provide api_key; Vertex can be configured via env vars or config.extra
        if config.api_key:
            self._client = genai.Client(api_key=config.api_key).aio
        else:
            # supports Vertex AI if env vars are set as per Google docs
            extra = config.extra or {}
            if extra.get("vertexai"):
                self._client = genai.Client(
                    vertexai=True,
                    project=extra.get("project"),
                    location=extra.get("location", "us-central1"),
                ).aio
            else:
                self._client = genai.Client().aio

    @staticmethod
    def _system_and_user(messages: List[LLMMessage]) -> tuple[str, str]:
        system_parts = [m.content for m in messages if m.role == "system" and m.content]
        user_parts = [m.content for m in messages if m.role == "user" and m.content]
        return "\n\n".join(system_parts).strip(), "\n\n".join(user_parts).strip()

    async def generate(self, messages: List[LLMMessage], **kwargs: Any) -> LLMResponse:
        system, user = self._system_and_user(messages)
        cfg = self._types.GenerateContentConfig(
            system_instruction=system or None,
            temperature=kwargs.pop("temperature", None),
            max_output_tokens=kwargs.pop("max_output_tokens", None),
        )
        resp = await self._client.models.generate_content(
            model=self.config.model,
            contents=user,
            config=cfg,
            **kwargs,
        )
        text = getattr(resp, "text", None) or ""
        return LLMResponse(text=text, raw={})



from __future__ import annotations
from typing import Any, Dict, List, Optional, Protocol
from pydantic import BaseModel, Field

Json = Dict[str, Any]

class LLMMessage(BaseModel):
    role: str
    content: str
    meta: Json = Field(default_factory=dict)

class LLMResponse(BaseModel):
    text: str
    raw: Json = Field(default_factory=dict)

class ProviderConfig(BaseModel):
    provider: str
    model: str
    api_key: Optional[str] = None
    base_url: Optional[str] = None
    timeout_s: float = 60.0
    extra: Json = Field(default_factory=dict)

class Provider(Protocol):
    config: ProviderConfig
    async def generate(self, messages: List[LLMMessage], **kwargs: Any) -> LLMResponse: ...



import sys
from pathlib import Path

# Ensure src/ layout imports work when running tests without editable install
ROOT = Path(__file__).resolve().parents[1]
SRC = ROOT / "src"
if str(SRC) not in sys.path:
    sys.path.insert(0, str(SRC))



from __future__ import annotations
from fastapi import FastAPI
from pydantic import BaseModel
from ..core.bus import CognitiveBus
from ..core.context import ContextObject

app = FastAPI(title="ADI Agentic AGI Service")

bus = CognitiveBus()

class Goal(BaseModel):
    text: str

@app.on_event("startup")
async def start_bus():
    await bus.start()

@app.post("/goal")
async def submit_goal(goal: Goal):
    ev = await bus.emit(topic="goal", type="TASK", sender="api", context=ContextObject(text=goal.text))
    return {"event_id": ev.event_id}

@app.get("/events")
async def list_events():
    return [e.model_dump() for e in bus.snapshot()]



from __future__ import annotations
import logging

def get_logger(name: str = "adi-agentic-agi") -> logging.Logger:
    logger = logging.getLogger(name)
    if not logger.handlers:
        h = logging.StreamHandler()
        fmt = logging.Formatter("%(asctime)s %(levelname)s %(name)s - %(message)s")
        h.setFormatter(fmt)
        logger.addHandler(h)
        logger.setLevel(logging.INFO)
    return logger



from __future__ import annotations
import asyncio, textwrap
from dataclasses import dataclass
from typing import Any, Dict, Optional

@dataclass
class SandboxResult:
    ok: bool
    stdout: str = ""
    stderr: str = ""
    result_repr: str = ""

class PythonSandbox:
    def __init__(self, timeout_s: float = 5.0):
        self.timeout_s = timeout_s

    async def run(self, code: str, globals_dict: Optional[Dict[str, Any]] = None) -> SandboxResult:
        globals_dict = globals_dict or {}
        local_vars: Dict[str, Any] = {}
        wrapped = textwrap.dedent(code)

        def _exec():
            exec(wrapped, globals_dict, local_vars)  # noqa: S102

        try:
            await asyncio.wait_for(asyncio.to_thread(_exec), timeout=self.timeout_s)
            out = local_vars.get("result", None)
            return SandboxResult(ok=True, result_repr=repr(out))
        except Exception as e:
            return SandboxResult(ok=False, stderr=str(e))



from __future__ import annotations
from typing import Optional, Set
from ..core.agent import BaseAgent
from ..core.decorators import agent_role
from ..core.events import BusEvent

@agent_role("ethics_guard")
class EthicsGuardrailAgent(BaseAgent):
    def topics(self) -> Optional[Set[str]]:
        return None

    def filter_event(self, event: BusEvent) -> bool:
        return "high_stakes" in (event.tags or [])

    async def handle(self, event: BusEvent) -> None:
        await self.report(parent=event, topic="alert.guardrail",
                          text="High-stakes tag detected. Require HITL approval before execution.", confidence=0.9)



from __future__ import annotations
from typing import Callable, Dict, Type

_AGENT_REGISTRY: Dict[str, Type] = {}

def agent_role(name: str) -> Callable[[Type], Type]:
    def _wrap(cls: Type) -> Type:
        _AGENT_REGISTRY[name] = cls
        setattr(cls, "__adi_role__", name)
        return cls
    return _wrap

def list_roles() -> Dict[str, Type]:
    return dict(_AGENT_REGISTRY)



import pytest, asyncio
from adi_agentic_agi.core.bus import CognitiveBus
from adi_agentic_agi.core.context import ContextObject

@pytest.mark.asyncio
async def test_bus_publish_and_subscribe():
    bus = CognitiveBus()
    got = []
    async def cb(ev):
        got.append(ev)
    bus.subscribe(agent_id="a1", callback=cb, topics={"t1"})
    await bus.start()
    await bus.emit(topic="t1", type="TASK", sender="u", context=ContextObject(text="x"), priority=5)
    await asyncio.sleep(0.05)
    await bus.stop()
    assert len(got) == 1
    assert got[0].topic == "t1"



from .registry import ToolRegistry, ToolSpec
from .openapi_loader import load_openapi_operations, OpenAPIOperation
from .openapi_binder import bind_openapi_directory
from .http_executor import OpenAPIHttpExecutor, HttpAuth
from .python_sandbox import PythonSandbox, SandboxResult

__all__ = [
  "ToolRegistry",
  "ToolSpec",
  "load_openapi_operations",
  "OpenAPIOperation",
  "bind_openapi_directory",
  "OpenAPIHttpExecutor",
  "HttpAuth",
  "PythonSandbox",
  "SandboxResult",
]



from __future__ import annotations
import os, json
from typing import List, Callable, Any

from .openapi_loader import load_openapi_operations, OpenAPIOperation
from .registry import ToolRegistry, ToolSpec
from .http_executor import OpenAPIHttpExecutor

def bind_openapi_directory(directory: str, registry: ToolRegistry, executor: OpenAPIHttpExecutor) -> List[str]:
    """Discover OpenAPI JSON specs and register *executable* tools bound to an HTTP executor."""
    added: List[str] = []
    for root, _, files in os.walk(directory):
        for f in files:
            if not f.lower().endswith(".json"):
                continue
            path = os.path.join(root, f)
            with open(path, "r", encoding="utf-8") as fp:
                spec = json.load(fp)

            for op in load_openapi_operations(spec):
                registry.register(_spec_for(op), _bound_tool(executor, op))
                added.append(op.tool_name)
    return added

def _spec_for(op: OpenAPIOperation) -> ToolSpec:
    return ToolSpec(
        name=op.tool_name,
        description=op.description,
        input_schema=op.input_schema,
        output_schema=op.output_schema,
        tags=["openapi", op.method.lower()],
    )

def _bound_tool(executor: OpenAPIHttpExecutor, op: OpenAPIOperation) -> Callable[..., Any]:
    async def _fn(**kwargs: Any) -> Any:
        return await executor.call(op, **kwargs)
    return _fn



from __future__ import annotations
from typing import Optional, List
from ..core.events import BusEvent
from .store import Checkpoint

class SQLEventStore:
    def __init__(self, db_url: str):
        self.db_url = db_url
        try:
            import sqlalchemy  # type: ignore
        except Exception as e:  # pragma: no cover
            raise RuntimeError("sql extra not installed. pip install adi-agentic-agi[sql]") from e

    async def put_event(self, event: BusEvent) -> None:
        raise NotImplementedError("Implement SQL persistence with SQLAlchemy async engine.")

    async def get_event(self, event_id: str) -> Optional[BusEvent]:
        raise NotImplementedError

    async def list_recent(self, limit: int = 100) -> List[BusEvent]:
        raise NotImplementedError

    async def checkpoint(self, notes: str = "") -> Checkpoint:
        import time, uuid
        return Checkpoint(checkpoint_id=str(uuid.uuid4()), ts=time.time(), notes=notes)



from __future__ import annotations
import asyncio
from typing import Any, Dict
from pydantic import BaseModel, Field

class HITLDecision(BaseModel):
    approved: bool
    notes: str = ""
    meta: Dict[str, Any] = Field(default_factory=dict)

class HITLGate:
    def __init__(self):
        loop = asyncio.get_event_loop()
        self._future: asyncio.Future[HITLDecision] = loop.create_future()

    def approve(self, notes: str = "") -> None:
        if not self._future.done():
            self._future.set_result(HITLDecision(approved=True, notes=notes))

    def reject(self, notes: str = "") -> None:
        if not self._future.done():
            self._future.set_result(HITLDecision(approved=False, notes=notes))

    async def wait(self, timeout_s: float = 3600) -> HITLDecision:
        return await asyncio.wait_for(self._future, timeout=timeout_s)



from __future__ import annotations
from typing import Any, Awaitable, Callable, Dict, Optional
from pydantic import BaseModel, Field

ToolFn = Callable[..., Awaitable[Any]]

class ToolSpec(BaseModel):
    name: str
    description: str = ""
    input_schema: Dict[str, Any] = Field(default_factory=dict)
    output_schema: Dict[str, Any] = Field(default_factory=dict)
    tags: list[str] = Field(default_factory=list)

class Tool(BaseModel):
    spec: ToolSpec
    fn: ToolFn

class ToolRegistry:
    def __init__(self):
        self._tools: Dict[str, Tool] = {}

    def register(self, spec: ToolSpec, fn: ToolFn) -> None:
        self._tools[spec.name] = Tool(spec=spec, fn=fn)

    def get(self, name: str) -> Optional[Tool]:
        return self._tools.get(name)

    def list_specs(self) -> Dict[str, ToolSpec]:
        return {k: v.spec for k, v in self._tools.items()}



from __future__ import annotations
from typing import Optional, Set
from ..core.agent import BaseAgent
from ..core.decorators import agent_role
from ..core.events import BusEvent

@agent_role("critic")
class CriticRefinerAgent(BaseAgent):
    def topics(self) -> Optional[Set[str]]:
        return {"result.draft"}

    def filter_event(self, event: BusEvent) -> bool:
        return event.topic == "result.draft" and event.type == "RESULT"

    async def handle(self, event: BusEvent) -> None:
        draft = event.context.text or ""
        score = 0.85 if len(draft) > 200 else 0.55
        await self.emit_confidence(parent=event, score=score, rationale="length/coverage heuristic")
        if score < 0.75:
            await self.report(parent=event, topic="task.refine",
                              text="Refine draft: add error handling, tests, clearer API boundaries.", confidence=0.7)
        else:
            await self.report(parent=event, topic="result.final",
                              text="Draft looks solid. Proceed to finalize and package.", confidence=0.85)



from __future__ import annotations
import re
from typing import Any, Dict, Optional
import httpx

from .openapi_loader import OpenAPIOperation

_PATH_PARAM = re.compile(r"{([^}]+)}")

class HttpAuth:
    def __init__(self, bearer_token: Optional[str] = None, headers: Optional[Dict[str, str]] = None):
        self.bearer_token = bearer_token
        self.headers = headers or {}

    def apply(self, headers: Dict[str, str]) -> Dict[str, str]:
        out = dict(headers)
        out.update(self.headers)
        if self.bearer_token and "authorization" not in {k.lower(): v for k, v in out.items()}:
            out["Authorization"] = f"Bearer {self.bearer_token}"
        return out

class OpenAPIHttpExecutor:
    """Executes OpenAPI-discovered operations over HTTP using httpx.AsyncClient."""
    def __init__(
        self,
        *,
        base_url: str,
        auth: Optional[HttpAuth] = None,
        timeout_s: float = 60.0,
        verify_tls: bool = True,
    ):
        self.base_url = base_url.rstrip("/")
        self.auth = auth or HttpAuth()
        self._client = httpx.AsyncClient(timeout=timeout_s, verify=verify_tls)

    async def aclose(self) -> None:
        await self._client.aclose()

    def _render_path(self, path: str, params: Dict[str, Any]) -> str:
        def repl(m: re.Match[str]) -> str:
            key = m.group(1)
            if key not in params:
                raise ValueError(f"Missing path param: {key}")
            return str(params.pop(key))
        return _PATH_PARAM.sub(repl, path)

    async def call(self, op: OpenAPIOperation, **kwargs: Any) -> Any:
        params = dict(kwargs)
        path = self._render_path(op.path, params)
        url = f"{self.base_url}{path}"

        # heuristics: if method is GET/DELETE -> query params, else JSON body + remaining query if 'query_' prefixed
        headers = self.auth.apply({"Accept": "application/json"})
        json_body: Optional[Dict[str, Any]] = None
        query: Dict[str, Any] = {}

        if op.method in {"GET", "DELETE"}:
            query = params
        else:
            # allow explicit query_ prefix
            query = {k[6:]: v for k, v in list(params.items()) if k.startswith("query_")}
            for k in list(params.keys()):
                if k.startswith("query_"):
                    params.pop(k)
            json_body = params or None
            headers["Content-Type"] = "application/json"

        r = await self._client.request(op.method, url, params=query or None, json=json_body, headers=headers)
        r.raise_for_status()
        ctype = r.headers.get("content-type", "")
        if "application/json" in ctype:
            return r.json()
        return r.text



from __future__ import annotations
import asyncio
import os
import shutil
import subprocess
import tempfile
from dataclasses import dataclass
from typing import Optional

@dataclass
class ContainerSandboxResult:
    ok: bool
    stdout: str = ""
    stderr: str = ""
    exit_code: int = 0

class DockerPythonSandbox:
    """Runs Python code in a locked-down Docker container.

    Security posture (best-effort; depends on host Docker settings):
    - No network: --network none
    - Read-only rootfs: --read-only
    - Drop all capabilities: --cap-drop ALL
    - Limited CPU/mem/pids
    - Mount a temp dir as /work (rw) for the script
    """
    def __init__(
        self,
        *,
        image: str = "python:3.11-slim",
        timeout_s: float = 8.0,
        mem: str = "256m",
        cpus: str = "1.0",
        pids_limit: int = 128,
    ):
        self.image = image
        self.timeout_s = timeout_s
        self.mem = mem
        self.cpus = cpus
        self.pids_limit = pids_limit
        self._docker = shutil.which("docker")

    async def run(self, code: str) -> ContainerSandboxResult:
        if not self._docker:
            raise RuntimeError("Docker not found. Install Docker and ensure `docker` is on PATH.")

        with tempfile.TemporaryDirectory(prefix="adi_agi_sbox_") as td:
            script_path = os.path.join(td, "main.py")
            with open(script_path, "w", encoding="utf-8") as f:
                f.write(code)

            cmd = [
                self._docker, "run", "--rm",
                "--network", "none",
                "--read-only",
                "--cap-drop", "ALL",
                "--pids-limit", str(self.pids_limit),
                "--memory", self.mem,
                "--cpus", self.cpus,
                "-v", f"{td}:/work:rw",
                "-w", "/work",
                self.image,
                "python", "main.py",
            ]

            def _call():
                return subprocess.run(cmd, capture_output=True, text=True)

            try:
                proc = await asyncio.wait_for(asyncio.to_thread(_call), timeout=self.timeout_s)
                return ContainerSandboxResult(
                    ok=proc.returncode == 0,
                    stdout=proc.stdout,
                    stderr=proc.stderr,
                    exit_code=proc.returncode,
                )
            except asyncio.TimeoutError:
                return ContainerSandboxResult(ok=False, stderr="Timeout", exit_code=124)



from __future__ import annotations
import asyncio
import contextlib
import time
import uuid
from dataclasses import dataclass
from typing import Awaitable, Callable, Dict, List, Optional, Set, Tuple

from .events import BusEvent
from .context import ContextObject
from .scheduler import PriorityScheduler
from .routing import ConfidenceRouter
from ..memory.store import EventStore, InMemoryEventStore

SubscriberFn = Callable[[BusEvent], Awaitable[None]]
FilterFn = Callable[[BusEvent], bool]

@dataclass(frozen=True)
class Subscription:
    agent_id: str
    filter_fn: FilterFn
    callback: SubscriberFn
    min_confidence: float = 0.0
    topics: Optional[Set[str]] = None

class CognitiveBus:
    """Async blackboard bus with:
    - priority scheduling + backpressure
    - policy routing (confidence-based)
    - optional persistence store for replay/checkpoints
    """

    def __init__(
        self,
        max_queue: int = 10_000,
        router: Optional[ConfidenceRouter] = None,
        store: Optional[EventStore] = None,
    ):
        self._subs: List[Subscription] = []
        self._running = False

        self._scheduler = PriorityScheduler(maxsize=max_queue)
        self._scheduler.set_handler(self._dispatch)

        self._store_mem: Dict[str, BusEvent] = {}
        self.router = router or ConfidenceRouter()
        self.store: EventStore = store or InMemoryEventStore()

    def snapshot(self) -> List[BusEvent]:
        return list(self._store_mem.values())

    def get(self, event_id: str) -> Optional[BusEvent]:
        return self._store_mem.get(event_id)

    def subscribe(
        self,
        *,
        agent_id: str,
        callback: SubscriberFn,
        filter_fn: Optional[FilterFn] = None,
        topics: Optional[Set[str]] = None,
        min_confidence: float = 0.0,
    ) -> None:
        if filter_fn is None:
            filter_fn = lambda e: True
        self._subs.append(Subscription(agent_id, filter_fn, callback, min_confidence, topics))

    async def publish(self, event: BusEvent) -> None:
        # apply routing policy before persistence/dispatch
        decision = self.router.decide(event)
        if decision.priority is not None:
            event.priority = decision.priority
        if decision.tags_add:
            event.tags = list(dict.fromkeys([*(event.tags or []), *decision.tags_add]))
        if decision.intended_for is not None:
            event.intended_for = decision.intended_for

        self._store_mem[event.event_id] = event
        await self.store.put_event(event)
        await self._scheduler.put(event)

    async def emit(
        self,
        *,
        topic: str,
        type: str,
        sender: str,
        context: Optional[ContextObject] = None,
        parent_id: Optional[str] = None,
        intended_for: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        confidence: Optional[float] = None,
        priority: int = 5,
        meta: Optional[Dict] = None,
    ) -> BusEvent:
        event = BusEvent(
            event_id=str(uuid.uuid4()),
            topic=topic,
            type=type,
            sender=sender,
            parent_id=parent_id,
            intended_for=intended_for or [],
            tags=tags or [],
            confidence=confidence,
            priority=priority,
            meta=meta or {"ts": time.time()},
            context=context or ContextObject(),
        )
        await self.publish(event)
        return event

    async def start(self) -> None:
        if self._running:
            return
        self._running = True
        await self._scheduler.start()

    async def stop(self) -> None:
        self._running = False
        await self._scheduler.stop()

    async def replay_recent(self, limit: int = 200) -> List[BusEvent]:
        """Replay last N events from the configured EventStore into in-memory blackboard (no dispatch)."""
        events = await self.store.list_recent(limit=limit)
        for ev in events:
            self._store_mem[ev.event_id] = ev
        return events

    async def _dispatch(self, event: BusEvent) -> None:
        direct = set(event.intended_for or [])
        targets: List[Tuple[float, Subscription]] = []

        for sub in self._subs:
            if sub.topics and event.topic not in sub.topics:
                continue
            if direct and sub.agent_id not in direct:
                continue
            if event.confidence is not None and event.confidence < sub.min_confidence:
                continue
            if sub.filter_fn(event):
                rank = float(event.confidence or 0.0)
                targets.append((rank, sub))

        targets.sort(key=lambda x: x[0], reverse=True)
        await asyncio.gather(*(sub.callback(event) for _, sub in targets), return_exceptions=True)



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from .events import BusEvent

@dataclass(frozen=True)
class RoutingDecision:
    intended_for: Optional[list[str]] = None
    priority: Optional[int] = None
    tags_add: Optional[list[str]] = None

class ConfidenceRouter:
    """Simple policy engine: route/escalate based on confidence + tags.

    Extend/replace this class in enterprise deployments.
    """

    def __init__(self, low_conf_threshold: float = 0.6, high_stakes_tag: str = "high_stakes"):
        self.low_conf_threshold = low_conf_threshold
        self.high_stakes_tag = high_stakes_tag

    def decide(self, event: BusEvent) -> RoutingDecision:
        # If low confidence RESULT -> escalate to critic by tagging
        if event.type == "RESULT" and event.confidence is not None and event.confidence < self.low_conf_threshold:
            return RoutingDecision(tags_add=["needs_review"], priority=max(1, event.priority - 1))
        # If high stakes -> require HITL tag for guardrail agent
        if self.high_stakes_tag in (event.tags or []):
            return RoutingDecision(tags_add=["needs_hitl"], priority=1)
        return RoutingDecision()



from __future__ import annotations
from typing import Optional, Set
from ..core.agent import BaseAgent
from ..core.decorators import agent_role
from ..core.events import BusEvent

@agent_role("retrospective")
class RetrospectiveAgent(BaseAgent):
    def topics(self) -> Optional[Set[str]]:
        return {"signal.failure"}

    def filter_event(self, event: BusEvent) -> bool:
        return event.topic == "signal.failure" and event.type == "ALERT"

    async def handle(self, event: BusEvent) -> None:
        await self.report(parent=event, topic="signal.prompt_patch",
                          text="Patch suggestion: Add explicit validation + structured output for next attempt.", confidence=0.8)



from __future__ import annotations
from typing import Any, List, Optional

from .base import ProviderConfig, LLMMessage, LLMResponse

class OpenAIProvider:
    """OpenAI Provider (real implementation).

    Uses the official `openai` Python SDK and the **Responses API** for text generation.
    """
    def __init__(self, config: ProviderConfig):
        self.config = config
        try:
            from openai import AsyncOpenAI  # type: ignore
        except Exception as e:  # pragma: no cover
            raise RuntimeError("Install provider extra: pip install adi-agentic-agi[openai]") from e

        kwargs: dict[str, Any] = {}
        if config.api_key:
            kwargs["api_key"] = config.api_key
        if config.base_url:
            kwargs["base_url"] = config.base_url
        # openai SDK supports default timeout/retries via httpx, but keep minimal here
        self._client = AsyncOpenAI(**kwargs)

    @staticmethod
    def _system_and_user(messages: List[LLMMessage]) -> tuple[str, str]:
        system_parts = [m.content for m in messages if m.role == "system" and m.content]
        user_parts = [m.content for m in messages if m.role in {"user", "developer"} and m.content]
        # keep it simple for now: single user string
        system = "\n\n".join(system_parts).strip()
        user = "\n\n".join(user_parts).strip()
        return system, user

    async def generate(self, messages: List[LLMMessage], **kwargs: Any) -> LLMResponse:
        system, user = self._system_and_user(messages)
        # prefer Responses API
        resp = await self._client.responses.create(
            model=self.config.model,
            instructions=system or None,
            input=user,
            **kwargs,
        )
        # SDK exposes `output_text`
        text = getattr(resp, "output_text", None) or ""
        return LLMResponse(text=text, raw={"id": getattr(resp, "id", None)})



from adi_agentic_agi.core.decorators import list_roles
from adi_agentic_agi.council import orchestrator, critic, tool_smith, context_librarian, ethics_guard, retrospective  # noqa: F401

def test_roles_registered():
    roles = list_roles()
    assert "orchestrator" in roles
    assert "critic" in roles
    assert "tool_smith" in roles
    assert "context_librarian" in roles



from __future__ import annotations
from typing import Optional, Set
from pydantic import BaseModel
from ..core.agent import BaseAgent
from ..core.decorators import agent_role
from ..core.events import BusEvent
from ..core.context import ContextObject

class Plan(BaseModel):
    research_question: str
    deliverable: str

@agent_role("orchestrator")
class OrchestratorAgent(BaseAgent):
    def topics(self) -> Optional[Set[str]]:
        return {"goal"}

    def filter_event(self, event: BusEvent) -> bool:
        return event.topic == "goal" and event.type == "TASK"

    async def handle(self, event: BusEvent) -> None:
        goal = event.context.text or ""
        plan = Plan(
            research_question=f"Research key points, constraints, and pitfalls for: {goal}",
            deliverable="Write production-grade Python module implementing solution with tests.",
        )
        await self.bus.emit(topic="task.research", type="TASK", sender=self.config.agent_id,
                            parent_id=event.event_id, context=ContextObject(text=plan.research_question), priority=2)
        await self.bus.emit(topic="task.code", type="TASK", sender=self.config.agent_id,
                            parent_id=event.event_id, context=ContextObject(text=plan.deliverable), priority=3)



from __future__ import annotations
from typing import List, Dict, Any

class RAGClient:
    def __init__(self, backend: str = "stub", **kwargs: Any):
        self.backend = backend
        self.kwargs = kwargs

    async def retrieve(self, query: str, k: int = 5) -> List[Dict[str, Any]]:
        return [{"text": "(stub) relevant context", "score": 0.0, "meta": {}}]



from __future__ import annotations
from typing import Optional, List
from ..core.events import BusEvent
from .store import Checkpoint

class RedisEventStore:
    def __init__(self, redis_url: str):
        self.redis_url = redis_url
        try:
            import redis  # type: ignore
        except Exception as e:  # pragma: no cover
            raise RuntimeError("redis extra not installed. pip install adi-agentic-agi[redis]") from e
        self._redis = redis.Redis.from_url(redis_url)

    async def put_event(self, event: BusEvent) -> None:
        self._redis.hset("events", event.event_id, event.model_dump_json())
        self._redis.rpush("event_order", event.event_id)

    async def get_event(self, event_id: str) -> Optional[BusEvent]:
        raw = self._redis.hget("events", event_id)
        if not raw:
            return None
        return BusEvent.model_validate_json(raw)

    async def list_recent(self, limit: int = 100) -> List[BusEvent]:
        ids = self._redis.lrange("event_order", -limit, -1)
        out: List[BusEvent] = []
        for i in ids:
            ev = await self.get_event(i.decode("utf-8"))
            if ev:
                out.append(ev)
        return out

    async def checkpoint(self, notes: str = "") -> Checkpoint:
        import time, uuid
        return Checkpoint(checkpoint_id=str(uuid.uuid4()), ts=time.time(), notes=notes)



from __future__ import annotations
from typing import Optional
from .store import EventStore, Checkpoint

class CheckpointManager:
    def __init__(self, store: EventStore):
        self.store = store

    async def snapshot(self, notes: str = "") -> Checkpoint:
        return await self.store.checkpoint(notes=notes)



from __future__ import annotations
from typing import Any, List

from .base import ProviderConfig, LLMMessage, LLMResponse

class AnthropicProvider:
    """Anthropic Provider (real implementation) using official `anthropic` SDK."""
    def __init__(self, config: ProviderConfig):
        self.config = config
        try:
            from anthropic import AsyncAnthropic  # type: ignore
        except Exception as e:  # pragma: no cover
            raise RuntimeError("Install provider extra: pip install adi-agentic-agi[anthropic]") from e

        kwargs: dict[str, Any] = {}
        if config.api_key:
            kwargs["api_key"] = config.api_key
        if config.base_url:
            kwargs["base_url"] = config.base_url
        self._client = AsyncAnthropic(**kwargs)

    @staticmethod
    def _system_and_messages(messages: List[LLMMessage]) -> tuple[str, list[dict[str, str]]]:
        system_parts = [m.content for m in messages if m.role == "system" and m.content]
        user_msgs: list[dict[str, str]] = []
        for m in messages:
            if m.role in {"user", "assistant"} and m.content:
                user_msgs.append({"role": m.role, "content": m.content})
        system = "\n\n".join(system_parts).strip()
        return system, user_msgs

    async def generate(self, messages: List[LLMMessage], **kwargs: Any) -> LLMResponse:
        system, msgs = self._system_and_messages(messages)
        max_tokens = int(kwargs.pop("max_tokens", self.config.extra.get("max_tokens", 1024)))
        resp = await self._client.messages.create(
            model=self.config.model,
            system=system or None,
            messages=msgs,
            max_tokens=max_tokens,
            **kwargs,
        )
        # resp.content is list of blocks; take text blocks
        parts = []
        for block in getattr(resp, "content", []) or []:
            txt = getattr(block, "text", None)
            if txt:
                parts.append(txt)
        return LLMResponse(text="\n".join(parts).strip(), raw={"id": getattr(resp, "id", None)})



from __future__ import annotations
from typing import Callable, TypeVar, Awaitable
from tenacity import retry, stop_after_attempt, wait_exponential

T = TypeVar("T")

def async_retry(attempts: int = 3, min_s: float = 0.2, max_s: float = 3.0):
    """Decorator for retrying async functions."""
    return retry(
        reraise=True,
        stop=stop_after_attempt(attempts),
        wait=wait_exponential(min=min_s, max=max_s),
    )



from __future__ import annotations
import abc
import asyncio
from typing import Any, Dict, Optional, Set, Callable, Awaitable

from pydantic import BaseModel, Field

from .bus import CognitiveBus
from .events import BusEvent
from .confidence import ConfidenceReport
from .context import ContextObject
from .retry import async_retry
from ..providers.base import Provider, LLMMessage

class AgentConfig(BaseModel):
    agent_id: str
    system_prompt: str = "You are a helpful agent."
    max_concurrency: int = 8
    min_confidence_to_claim: float = 0.0
    retry_attempts: int = 1
    extra: Dict[str, Any] = Field(default_factory=dict)

class BaseAgent(abc.ABC):
    def __init__(self, *, config: AgentConfig, bus: CognitiveBus, provider: Optional[Provider] = None):
        self.config = config
        self.bus = bus
        self.provider = provider
        self._sem = asyncio.Semaphore(config.max_concurrency)

    def subscribe(self) -> None:
        self.bus.subscribe(
            agent_id=self.config.agent_id,
            callback=self._guarded_handle,
            filter_fn=self.filter_event,
            min_confidence=self.config.min_confidence_to_claim,
            topics=self.topics(),
        )

    def topics(self) -> Optional[Set[str]]:
        return None

    def filter_event(self, event: BusEvent) -> bool:
        return True

    async def _guarded_handle(self, event: BusEvent) -> None:
        async with self._sem:
            await self._handle_with_retry(event)

    async def _handle_with_retry(self, event: BusEvent) -> None:
        if self.config.retry_attempts <= 1:
            await self.handle(event)
            return

        @async_retry(attempts=self.config.retry_attempts)
        async def _run() -> None:
            await self.handle(event)

        try:
            await _run()
        except Exception as e:
            # publish failure signal for retrospective learning
            await self.bus.emit(
                topic="signal.failure",
                type="ALERT",
                sender=self.config.agent_id,
                parent_id=event.event_id,
                context=ContextObject(text=f"Agent failed after retries: {e}"),
                confidence=0.0,
                priority=1,
            )

    @abc.abstractmethod
    async def handle(self, event: BusEvent) -> None:
        ...

    async def think(self, user_text: str, *, extra_system: Optional[str] = None, **kwargs: Any) -> str:
        if not self.provider:
            raise RuntimeError(f"{self.config.agent_id} has no provider attached.")
        system = self.config.system_prompt + (("\n" + extra_system) if extra_system else "")
        messages = [LLMMessage(role="system", content=system), LLMMessage(role="user", content=user_text)]
        resp = await self.provider.generate(messages, **kwargs)
        return resp.text

    async def emit_confidence(self, *, parent: BusEvent, score: float, rationale: str = "") -> None:
        await self.bus.emit(
            topic="signal.confidence",
            type="THOUGHT",
            sender=self.config.agent_id,
            parent_id=parent.event_id,
            confidence=score,
            meta={"rationale": rationale, "source_event_topic": parent.topic},
        )

    async def report(self, *, parent: BusEvent, topic: str, text: str, confidence: float = 0.7) -> None:
        await self.bus.emit(
            topic=topic,
            type="RESULT",
            sender=self.config.agent_id,
            parent_id=parent.event_id,
            context=ContextObject(text=text),
            confidence=confidence,
        )

    async def score_fit(self, event: BusEvent) -> ConfidenceReport:
        return ConfidenceReport(score=0.5, rationale="default")



import pytest
import httpx

from adi_agentic_agi.tools.openapi_binder import bind_openapi_directory
from adi_agentic_agi.tools.http_executor import OpenAPIHttpExecutor, HttpAuth
from adi_agentic_agi.tools.registry import ToolRegistry

@pytest.mark.asyncio
async def test_openapi_binder_executes(tmp_path):
    # create minimal OpenAPI spec
    spec = {
        "openapi": "3.0.0",
        "info": {"title": "T", "version": "1.0"},
        "paths": {
            "/echo": {
                "post": {
                    "operationId": "echo",
                    "requestBody": {
                        "content": {"application/json": {"schema": {"type": "object"}}}
                    },
                    "responses": {
                        "200": {
                            "description": "ok",
                            "content": {"application/json": {"schema": {"type": "object"}}},
                        }
                    },
                }
            }
        },
    }
    d = tmp_path / "specs"
    d.mkdir()
    (d / "api.json").write_text(__import__("json").dumps(spec), encoding="utf-8")

    # mock httpx by monkeypatching executor client transport
    def handler(request: httpx.Request) -> httpx.Response:
        assert request.url.path == "/echo"
        return httpx.Response(200, json={"ok": True})

    transport = httpx.MockTransport(handler)
    ex = OpenAPIHttpExecutor(base_url="https://example.com", auth=HttpAuth())
    # swap internal client
    ex._client = httpx.AsyncClient(transport=transport, base_url="https://example.com")  # type: ignore[attr-defined]

    reg = ToolRegistry()
    bind_openapi_directory(str(d), reg, ex)
    tool = reg.get("echo")
    assert tool is not None
    out = await tool.fn(a=1)
    assert out["ok"] is True
    await ex.aclose()



from __future__ import annotations
from typing import Any, List
from .base import ProviderConfig, LLMMessage, LLMResponse

class LocalProvider:
    def __init__(self, config: ProviderConfig):
        self.config = config

    async def generate(self, messages: List[LLMMessage], **kwargs: Any) -> LLMResponse:
        last = next((m.content for m in reversed(messages) if m.role == "user"), "")
        return LLMResponse(text=f"[LOCAL:{self.config.model}] {last}", raw={"stub": True})



from __future__ import annotations
from typing import Any, Dict, List
from pydantic import BaseModel, Field

Json = Dict[str, Any]

class OpenAPIOperation(BaseModel):
    tool_name: str
    method: str
    path: str
    description: str = ""
    input_schema: Json = Field(default_factory=dict)
    output_schema: Json = Field(default_factory=dict)

def _normalize_tool_name(method: str, path: str) -> str:
    name = f"{method.lower()}_{path.strip('/').replace('/', '_').replace('{', '').replace('}', '')}"
    return name or "op"

def load_openapi_operations(spec: Json) -> List[OpenAPIOperation]:
    ops: List[OpenAPIOperation] = []
    paths = spec.get("paths", {}) or {}
    for path, methods in paths.items():
        for method, op in (methods or {}).items():
            if method.lower() not in {"get", "post", "put", "patch", "delete"}:
                continue
            tool_name = op.get("operationId") or _normalize_tool_name(method, path)
            desc = op.get("description") or op.get("summary") or ""
            req = (op.get("requestBody") or {}).get("content", {}) or {}
            input_schema = {}
            if req:
                app = req.get("application/json") or next(iter(req.values()), {})
                input_schema = (app.get("schema") or {}) if isinstance(app, dict) else {}
            resp = (op.get("responses") or {}).get("200") or (op.get("responses") or {}).get("201") or {}
            content = (resp.get("content") or {}) if isinstance(resp, dict) else {}
            output_schema = {}
            if content:
                app = content.get("application/json") or next(iter(content.values()), {})
                output_schema = (app.get("schema") or {}) if isinstance(app, dict) else {}
            ops.append(OpenAPIOperation(
                tool_name=tool_name,
                method=method.upper(),
                path=path,
                description=desc,
                input_schema=input_schema,
                output_schema=output_schema,
            ))
    return ops



from __future__ import annotations
import asyncio
import contextlib
from dataclasses import dataclass
from typing import Awaitable, Callable, Optional, Tuple

from .events import BusEvent

Handler = Callable[[BusEvent], Awaitable[None]]

@dataclass(order=True)
class _PQItem:
    priority: int
    seq: int
    event: BusEvent

class PriorityScheduler:
    """Async priority scheduler with backpressure and graceful shutdown."""

    def __init__(self, maxsize: int = 10_000):
        self._pq: asyncio.PriorityQueue[_PQItem] = asyncio.PriorityQueue(maxsize=maxsize)
        self._running = False
        self._task: Optional[asyncio.Task] = None
        self._seq = 0
        self._handler: Optional[Handler] = None

    def set_handler(self, handler: Handler) -> None:
        self._handler = handler

    async def put(self, event: BusEvent) -> None:
        # Lower number => higher priority
        self._seq += 1
        await self._pq.put(_PQItem(priority=int(event.priority), seq=self._seq, event=event))

    async def start(self) -> None:
        if self._running:
            return
        if self._handler is None:
            raise RuntimeError("PriorityScheduler handler not set.")
        self._running = True
        self._task = asyncio.create_task(self._loop())

    async def stop(self) -> None:
        self._running = False
        if self._task:
            self._task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._task

    async def _loop(self) -> None:
        assert self._handler is not None
        while self._running:
            item = await self._pq.get()
            await self._handler(item.event)



from ._version import __version__

from .core.bus import CognitiveBus
from .core.agent import BaseAgent, AgentConfig
from .core.context import ContextObject, MediaBlob
from .core.routing import ConfidenceRouter
from .providers.base import ProviderConfig
from .tools.http_executor import OpenAPIHttpExecutor, HttpAuth
from .tools.openapi_binder import bind_openapi_directory

__all__ = [
    "__version__",
    "CognitiveBus",
    "BaseAgent",
    "AgentConfig",
    "ContextObject",
    "MediaBlob",
    "ConfidenceRouter",
    "ProviderConfig",
    "OpenAPIHttpExecutor",
    "HttpAuth",
    "bind_openapi_directory",
]



from __future__ import annotations
import asyncio
from ..core.bus import CognitiveBus
from ..core.agent import AgentConfig
from ..providers.base import ProviderConfig
from ..providers.local_provider import LocalProvider
from ..council.orchestrator import OrchestratorAgent
from ..council.context_librarian import ContextLibrarianAgent
from ..council.tool_smith import ToolSmithAgent
from ..council.critic import CriticRefinerAgent
from ..council.ethics_guard import EthicsGuardrailAgent
from ..council.retrospective import RetrospectiveAgent
from ..core.context import ContextObject

async def run():
    bus = CognitiveBus()
    await bus.start()
    provider = LocalProvider(ProviderConfig(provider="local", model="stub-llm"))

    agents = [
        OrchestratorAgent(config=AgentConfig(agent_id="orchestrator", retry_attempts=2), bus=bus, provider=provider),
        ContextLibrarianAgent(config=AgentConfig(agent_id="librarian", retry_attempts=2), bus=bus, provider=provider),
        ToolSmithAgent(config=AgentConfig(agent_id="tool_smith", retry_attempts=1, extra={"tool_smith_container": False}),
                      bus=bus, provider=provider),
        CriticRefinerAgent(config=AgentConfig(agent_id="critic", retry_attempts=1), bus=bus, provider=provider),
        EthicsGuardrailAgent(config=AgentConfig(agent_id="ethics", retry_attempts=1), bus=bus, provider=provider),
        RetrospectiveAgent(config=AgentConfig(agent_id="retro", retry_attempts=1), bus=bus, provider=provider),
    ]
    for a in agents:
        a.subscribe()

    await bus.emit(
        topic="goal",
        type="TASK",
        sender="user",
        context=ContextObject(text="Design a caching layer for API responses with TTL + tests."),
        priority=1,
    )

    await asyncio.sleep(1.0)
    for e in bus.snapshot()[-25:]:
        t = (e.context.text or "").replace("\n", " ")
        if len(t) > 90:
            t = t[:90] + "..."
        print(f"{e.type:<6} {e.topic:<18} sender={e.sender} conf={e.confidence} prio={e.priority} text={t!r}")
    await bus.stop()

if __name__ == "__main__":
    asyncio.run(run())
